"""Demonstration of Interactive Clean running remotely, starting a remote webserver, and
the generated GUI shown locally automatically."""
import os
import sys
sys.path.insert(0, os.path.abspath('./'))
import casaremote
from cubevis.utils import find_ws_address

# This is the name of an 'ssh_ipykernel' remote jupyter kernel on your local machine
# See https://pypi.org/project/ssh-ipykernel/ for more details on creating this
remote_jupyter_kernel_name = ""

# Start the remote jupyter kernel and initialize the environment
_, backchannel_port = find_ws_address("localhost")
connect_info_file = casaremote.start_remote_kernel(remote_jupyter_kernel_name, False, backchannel_port=backchannel_port)
kc = casaremote.connect_remote(connect_info_file, "zuul07")

# Initialize the remote kernel environment for Interactive Clean
# TODO: Move this into its own initialization function? Or keep it more explicit here
casaremote.exe_and_print(kc, "import cubevis")
casaremote.exe_and_print(kc, "from cubevis import InteractiveClean")
casaremote.exe_and_print(kc, "import casaremote")
casaremote.exe_and_print(kc, "import asyncio")
casaremote.exe_and_print(kc, "t = casaremote.TestProc(" + str(backchannel_port)+", asyncio.get_running_loop())")
casaremote.exe_and_print(kc, "t.ic = None")

# Create TestProc object in top-level scope
casaremote.exe_and_print(kc, "t = casaremote.TestProc(" + str(backchannel_port)+", asyncio.get_running_loop())")
casaremote.exe_and_print(kc, "t.start_backchannel()")

# Initialize the remote Interactive Clean instance
ms_path = 'refim_point_withline.ms'
img = 'test'

def init_ic(kc, *args, **kwargs):
    delete_cmd = "os.system('rm -rf + " + kwargs['imagename'] + ".* *.html *.log')"
    casaremote.exe_and_print(kc, delete_cmd)
    cmd = "t.ic = InteractiveClean(" + casaremote.kwarg_string(*args, **kwargs) + ")"
    casaremote.exe_and_print(kc, cmd)

init_ic(kc,
        remote=True,
        serve_webpage=True,
        vis=ms_path,
        imagename=img,
        imsize=512,
        cell='12.0arcsec',
        specmode='cube',
        interpolation='nearest',
        nchan=5,
        start='1.0GHz',
        width='0.2GHz',
        pblimit=-1e-05,
        deconvolver='hogbom',
        threshold='0.001Jy',
        niter=50,
        cycleniter=10,
        cyclefactor=3,
        scales=[0,3,10] )

# Start Interactive Clean remotely
casaremote.exe_and_print(kc, "result = t.ic()")

print("Reconnection info file: ", connect_info_file)

# Query ports, auto port-forward, and open webpage
port_cmd = casaremote.create_cmd_json("self.get_ports")
port_cmd_id = casaremote.exe_remote_subproc(kc, port_cmd)
ports = casaremote.get_subproc_return_val(kc, port_cmd_id)
ssh_cmd = casaremote.build_ssh_tunnel("zuul07", ports)

print("Tunneling connections: " + ssh_cmd)
os.system(ssh_cmd)
os.system("open http://127.0.0.1:" + str(ports[-1]))

input("Press enter to exit")


########################################################################
#
# Copyright (C) 2023
# Associated Universities, Inc. Washington DC, USA.
#
# This script is free software; you can redistribute it and/or modify it
# under the terms of the GNU Library General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
# License for more details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#
# Correspondence concerning AIPS++ should be adressed as follows:
#        Internet email: casa-feedback@nrao.edu.
#        Postal address: AIPS++ Project Office
#                        National Radio Astronomy Observatory
#                        520 Edgemont Road
#                        Charlottesville, VA 22903-2475 USA
#
########################################################################
from bokeh.models import TickFormatter
from bokeh.util.compiler import TypeScript
from bokeh.core.properties import Instance, String
from cubevis.bokeh.sources import ImageDataSource
from ..state import casalib_url, cubevisjs_url
from .. import BokehInit

class WcsTicks(TickFormatter,BokehInit):

    ## which axis are we labeling
    axis = String( )

    ## source containing the WCS information
    image_source = Instance(ImageDataSource)

    ###################################################################
    ### filled from cubevis.bokeh.state._initialize._order_bokeh_js ###
    ###################################################################
    #__javascript__ = [ casalib_url( ), cubevisjs_url( ) ]

    def __init__( self, *args, **kwargs ):
        super( ).__init__( *args, **kwargs )



########################################################################
#
# Copyright (C) 2021,2022,2023,2025
# Associated Universities, Inc. Washington DC, USA.
#
# This script is free software; you can redistribute it and/or modify it
# under the terms of the GNU Library General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
# License for more details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#
# Correspondence concerning AIPS++ should be adressed as follows:
#        Internet email: casa-feedback@nrao.edu.
#        Postal address: AIPS++ Project Office
#                        National Radio Astronomy Observatory
#                        520 Edgemont Road
#                        Charlottesville, VA 22903-2475 USA
#
########################################################################
'''This contains functions to inject the ``cubevisjs`` library into the
generated HTML that is used to display the Bokeh plots that ``casagui``'s
applications produce'''
import os
import re
import logging
from os import path
from enum import Enum
from os.path import dirname, join, basename, abspath
from urllib.parse import urlparse
from bokeh import resources
from bokeh.settings import settings
from ...utils import path_to_url, static_vars, have_network

logger = logging.getLogger(__name__)

_CUBEVIS_LIBS = {}
# Selector for which source to prefer for JavaScript files. Bokeh's Python
# package comes with static versions of the Bokeh JavaScript libraries.
# cubevis libraries are available both as static versions contained within
# the Python package and from an NRAO website.
#
# None = auto-detect, 'local' = prefer local, 'network' = prefer network
class JsLoading(Enum):
    AUTO = 'auto'
    LOCAL = 'local'
    NETWORK = 'network'

_JS_STRATEGY = JsLoading.AUTO

# Package-level registry
_JUPYTER_STATE = {
    'models_registered': set()
}

def get_cubevis_libs( ):
    """Get the package-level default cubevis paths"""
    return _CUBEVIS_LIBS

def set_cubevis_lib(**kw):
    """Set custom paths for cubevis JavaScript libraries"""
    libs = get_cubevis_libs()
    for lib in ['bokeh', 'bokeh_widgets', 'bokeh_tables', 'casalib', 'cubevisjs']:
        if lib in kw:
            libs[lib] = kw[lib]
    return libs

def get_js_loading( ):
    """Get JavaScript loading strategy.
    """
    return _JS_STRATEGY

def set_js_loading(strategy):
    """Set JavaScript loading strategy.

    Args:
        strategy (JsLoading or str):
            - 'auto': Smart selection based on environment
            - 'local': Always prefer local static files when available
            - 'network': Always prefer network CDN when available
    """
    global _JS_STRATEGY
    if isinstance(strategy, str):
        _JS_STRATEGY = JsLoading(strategy)
    elif isinstance(strategy, JsLoading):
        _JS_STRATEGY = strategy
    else:
        raise TypeError("'strategy' must be a valid JsLoading string or enum member")

def get_js_loading_selection( ):
    # Check for user override
    global _JS_STRATEGY

    prefer_network = False
    prefer_local = False

    in_jupyter = is_jupyter()
    use_network = have_network()

    if _JS_STRATEGY is JsLoading.LOCAL:
        prefer_network = False
        prefer_local = True
        logger.debug("Strategy: User override -> prefer local")
    elif _JS_STRATEGY is JsLoading.NETWORK:
        prefer_network = True
        prefer_local = False
        logger.debug("Strategy: User override -> prefer network")
    else:
        # Auto-detect optimal strategy
        if in_jupyter and use_network:
            # Jupyter notebooks work best with CDN versions
            prefer_network = True
            prefer_local = False
            logger.debug("Strategy: Jupyter + network -> prefer CDN")
        elif not use_network:
            # No choice - must use local
            prefer_network = False
            prefer_local = True
            logger.debug("Strategy: No network -> local only")
        else:
            # Standalone application with network -> prefer local for speed/reliability
            prefer_network = False
            prefer_local = True
            logger.debug("Strategy: Standalone + network -> prefer local static")

    return prefer_local, prefer_network

def is_jupyter():
    """Check if running in Jupyter environment"""
    try:
        from IPython import get_ipython
        return get_ipython() is not None
    except ImportError:
        return False

def resolve_js_library_paths():
    """Resolve paths to all required JavaScript libraries based on environment and availability.

    Strategy:
    - Jupyter notebooks: Prefer network CDN (required for proper functionality)
    - Standalone apps with network: Prefer local static (faster, more reliable)
    - No network: Use local static (only option)
    - Corporate/restricted networks: Prefer local static (firewall bypass)

    Returns:
        dict: Dictionary with resolved paths for 'casalib', 'bokeh', 'bokeh_widgets',
              'bokeh_tables', and 'cubevisjs'
    """
    ### These functions also select remote/local based on Jupyter or
    ### no network because these URLs also are included in the
    ### __javascript__ variables for Bokeh model derived member variables.
    from . import casalib_url as get_casalib_url
    from . import cubevisjs_url as get_cubevisjs_url

    # Start with user-defined overrides
    result = get_cubevis_libs().copy()

    # Handle {JSLIB}/ substitution for user paths
    for key, value in result.items():
        if value and value.startswith("{JSLIB}/"):
            result[key] = path_to_url(value[8:])

    prefer_local, prefer_network = get_js_loading_selection( )

    # Get available sources
    bokeh_static_js = {}
    bokeh_static_available = False

    try:
        bokeh_static_path = settings.bokehjs_path()
        js_dir = join(bokeh_static_path, 'js')
        logger.debug( f"Bokeh static pat: {js_dir}" )

        if path.exists(js_dir):
            for filename in os.listdir(js_dir):
                if filename.endswith('.js'):
                    file_path = f"file://{abspath(join(js_dir, filename))}"

                    ## js_dir seems to contain both minified and non minified versions of
                    ## the bokeh librarys. If we ever want to allow for non-minified
                    ## versions, e.g. for debugging, we'll need to look here
                    if re.match(r'.*bokeh(?:-\d+\.\d+\.\d+)?\.min\.js$', filename):
                        bokeh_static_js['bokeh'] = file_path
                    elif re.match(r'.*bokeh-widgets(?:-\d+\.\d+\.\d+)?\.min\.js$', filename):
                        bokeh_static_js['bokeh_widgets'] = file_path
                    elif re.match(r'.*bokeh-tables(?:-\d+\.\d+\.\d+)?\.min\.js$', filename):
                        bokeh_static_js['bokeh_tables'] = file_path

            bokeh_static_available = len(bokeh_static_js) > 0

    except Exception as e:
        logger.warning(f"Could not access Bokeh static files: {e}")

    # Network connection available
    use_network = have_network()

    # Fill in missing Bokeh libraries using the selected strategy
    for lib in ['bokeh', 'bokeh_widgets', 'bokeh_tables']:
        if lib not in result or result[lib] is None:
            if prefer_local and bokeh_static_available and lib in bokeh_static_js:
                result[lib] = bokeh_static_js[lib]
                logger.debug(f"Using local static for {lib}")
            elif prefer_network and use_network:
                # Let Bokeh handle CDN URLs - return None to use defaults
                result[lib] = None
                logger.debug(f"Using CDN for {lib}")
            elif bokeh_static_available and lib in bokeh_static_js:
                # Fallback to local if network preference fails
                result[lib] = bokeh_static_js[lib]
                logger.debug(f"Fallback to local static for {lib}")
            else:
                # Last resort - try local cache
                try:
                    cached_path = path_to_url(f"{lib}.min.js")
                    if cached_path != f"{lib}.min.js":  # Found in cache
                        result[lib] = cached_path
                        logger.debug(f"Using local cache for {lib}")
                    else:
                        result[lib] = None
                        logger.debug(f"No source found for {lib}")
                except:
                    result[lib] = None

    # Fill in cubevis-specific libraries
    if 'casalib' not in result or result['casalib'] is None:
        result['casalib'] = get_casalib_url( )

    if 'cubevisjs' not in result or result['cubevisjs'] is None:
        result['cubevisjs'] = get_cubevisjs_url( )

    logger.debug(f"Resolved JS library paths: {result}")
    logger.debug(f"Environment: jupyter={is_jupyter( )}, network={use_network}, strategy={'network' if prefer_network else 'local'}")

    return result

@static_vars(initialized=False)
def order_bokeh_js():
    """Initialize `bokeh` for use with the ``cubevisjs`` extensions.

    This function injects the cubevis JavaScript libraries into Bokeh's resource
    loading system in the proper order: casalib → bokeh core → bokeh widgets →
    bokeh tables → cubevisjs.

    The function automatically determines the best source for JavaScript libraries
    based on the current environment (Jupyter vs standalone, network availability, etc.).
    """

    if order_bokeh_js.initialized:
        return

    # Get all resolved library paths
    js_paths = resolve_js_library_paths()

    logger.debug(f"order_bokeh_js() initializing with paths: {js_paths}")
    order_bokeh_js.initialized = True

    # Store original Bokeh js_files function
    resources.Resources._old_js_files = resources.Resources.js_files

    def js_files(self):
        """Replacement function that returns JavaScript files in the correct order"""

        # Get original Bokeh URLs as fallback
        original_urls = resources.Resources._old_js_files.fget(self)
        logger.debug( f"Original Bokeh JavaScript: {original_urls}" )

        def get_url_or_fallback(lib_key, fallback_pattern=None):
            """Get resolved URL or find fallback from original Bokeh URLs"""
            resolved_path = js_paths.get(lib_key)

            if resolved_path:
                return resolved_path

            # Look for fallback in original URLs
            if fallback_pattern:
                for url in original_urls:
                    if re.match(fallback_pattern, url):
                        return url

            return None

        # Build the ordered list
        ordered_js = []

        # 1. casalib (always first)
        casalib_url = get_url_or_fallback('casalib')
        if casalib_url:
            ordered_js.append(casalib_url)

        # 2. bokeh core
        bokeh_url = get_url_or_fallback('bokeh', r'.*/bokeh-\d+\.\d+\.\d+(?:\.min)?\.js$')
        if bokeh_url:
            ordered_js.append(bokeh_url)

        # 3. bokeh widgets
        widgets_url = get_url_or_fallback('bokeh_widgets', r'.*/bokeh-widgets-\d+\.\d+\.\d+(?:\.min)?\.js$')
        if widgets_url:
            ordered_js.append(widgets_url)

        # 4. bokeh tables
        tables_url = get_url_or_fallback('bokeh_tables', r'.*/bokeh-tables-\d+\.\d+\.\d+(?:\.min)?\.js$')
        if tables_url:
            ordered_js.append(tables_url)

        # 5. cubevisjs (always last)
        cubevisjs_url = get_url_or_fallback('cubevisjs')
        if cubevisjs_url:
            ordered_js.append(cubevisjs_url)

        logger.debug(f"Final JS load order: {ordered_js}")
        return ordered_js

    # Replace Bokeh's js_files property
    resources.Resources.js_files = property(js_files)
    return

#def get_bokeh_js_paths( ):
#    modes = ['cdn','inline','server','server-dev','relative','relative-dev','absolute','absolute-dev']
#    return { 'new': { mode: resources.Resources(mode=mode).js_files for mode in modes },
#             'old': { mode: resources.Resources(mode=mode)._old_js_files for mode in modes } }
def get_bokeh_js_paths( ):
    return resources.Resources(mode='cdn').js_files

def get_jupyter_state( ):
    """Get the package-level Jupyter state"""
    return _JUPYTER_STATE

def register_model(model_class):
    """Register a model class that needs dependencies"""
    _JUPYTER_STATE['models_registered'].add(model_class.__name__)



""" Use the python logging interface for cubevis logs.
If casalog is available, then send log messages to it.
Otherwise, use the base python logging mechanisms.
"""
import logging

try:
	from casatasks import casalog
	__casalog_avail = True
except ImportError:
	__casalog_avail = False

class _Logging:
	# singleton instance
	logger = None
	log_handler = None

class CasalogHandler(logging.Handler):
	""" Logs to casalog, as appropriate """
	# map from python logging levels to casalog priorities
	level_to_priority = {
		"DEBUG": "DEBUG",
		"INFO": "INFO",
		"WARNING": "WARN",
		"ERROR": "SEVERE",
		"CRITICAL": "SEVERE"
	}

	def _level_to_levelname(self, level):
		if (level >= 50):
			return "CRITICAL"
		if (level >= 40):
			return "ERROR"
		if (level >= 30):
			return "WARNING"
		if (level >= 20):
			return "INFO"
		return "DEBUG"

	def setLevel(self, level_or_levelname):
		super().setLevel(level_or_levelname)
		if type(level_or_levelname) != str:
			levelname = self._level_to_levelname(level_or_levelname)
		else:
			levelname = level_or_levelname
		casalog.filter(CasalogHandler.level_to_priority[levelname])

	def emit(self, record):
		# promote log level for casalog
		levelname = self._level_to_levelname(_Logging.logger.getEffectiveLevel())
		if levelname == "DEBUG":
			casalog.filter(CasalogHandler.level_to_priority["DEBUG"])

		# get log properties
		origin = (record.filename) if (record.filename != None) else ("")
		casalog.origin(origin)
		priority = CasalogHandler.level_to_priority[record.levelname]
		self.format(record) # populates record.message

		# emit the log
		casalog.post(message=record.message, priority=priority, origin=record.funcName)

class CasalogLogger(logging.getLoggerClass()):
	def setLevel(self, level):
		super().setLevel(level)
		_Logging.log_handler.setLevel(level)

def get_logger():
	""" Returns the singleton logger instance. """
	if _Logging.logger != None:
		return _Logging.logger

	if __casalog_avail:
		_Logging.log_handler = CasalogHandler()

		curr_logger_class = logging.getLoggerClass()
		logging.setLoggerClass(CasalogLogger)
		_Logging.logger = logging.getLogger(__name__)
		logging.setLoggerClass(curr_logger_class)

		_Logging.logger.addHandler(_Logging.log_handler)
		_Logging.logger.propagate = False
	else:
		_Logging.logger = logging.getLogger(__name__)
	return _Logging.logger


########################################################################
#
# Copyright (C) 2024
# Associated Universities, Inc. Washington DC, USA.
#
# This script is free software; you can redistribute it and/or modify it
# under the terms of the GNU Library General Public License as published by
# the Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
# License for more details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#
# Correspondence concerning AIPS++ should be adressed as follows:
#        Internet email: casa-feedback@nrao.edu.
#        Postal address: AIPS++ Project Office
#                        National Radio Astronomy Observatory
#                        520 Edgemont Road
#                        Charlottesville, VA 22903-2475 USA
#
########################################################################
##################### generated by xml-casa (v2) from createmask.xml
##################### 26fdebc4cbd4e58e368f393b9b7ccde7 ##############################
from __future__ import absolute_import
from casashell.private.stack_manip import find_local as __sf__
from casashell.private.stack_manip import find_frame as _find_frame
from casatools.typecheck import validator as _pc
from casatools.coercetype import coerce as _coerce
from casatools.errors import create_error_string
from casatasks.private.task_logging import start_log as _start_log
from casatasks.private.task_logging import end_log as _end_log
from cubevis.private.apps import InteractiveClean
from collections import OrderedDict
import numpy
import sys
import os

import shutil

def static_var(varname, value):
    def decorate(func):
        setattr(func, varname, value)
        return func
    return decorate


def _createmask_t( *args, **kwargs ):
    ic = InteractiveClean( *args, **kwargs )
    return ic( )


class _createmask:
    """
    createmask ---- Create mask cubes interactively with a GUI

    Create a mask cube for one or more images using a GUI.
    This task is based upon the same tools that are used to build
    the interactive clean application so the same command keys are
    used for both applications.


    --------- parameter descriptions ---------------------------------------------

    image
    mask
    create

    --------- examples -----------------------------------------------------------




    """

    _info_group_ = """imaging"""
    _info_desc_ = """Create mask cubes interactively with a GUI"""

    __schema = { 'image': {'anyof': [{'type': 'cReqPath', 'coerce': _coerce.expand_path}, {'type': 'cReqPathVec', 'coerce': [_coerce.to_list,_coerce.expand_pathvec]}]}, 'mask': {'anyof': [{'type': 'cPath', 'coerce': _coerce.expand_path}, {'type': 'cPathVec', 'coerce': [_coerce.to_list,_coerce.expand_pathvec]}]}, 'create': {'type': 'cBool'},  }

    def __init__(self):
        self.__stdout = None
        self.__stderr = None
        self.__root_frame_ = None

    def __globals_(self):
        if self.__root_frame_ is None:
            self.__root_frame_ = _find_frame( )
            assert self.__root_frame_ is not None, "could not find CASAshell global frame"
        return self.__root_frame_

    def __to_string_(self,value):
        if type(value) is str:
            return "'%s'" % value
        else:
            return str(value)

    def __validate_(self,doc,schema):
        return _pc.validate(doc,schema)

    def __do_inp_output(self,param_prefix,description_str,formatting_chars):
        out = self.__stdout or sys.stdout
        description = description_str.split( )
        prefix_width = 23 + 23 + 4
        output = [ ]
        addon = ''
        first_addon = True
        if len(description) == 0:
            out.write(param_prefix + " #\n")
            return
        while len(description) > 0:
            ## starting a new line.....................................................................
            if len(output) == 0:
                ## for first line add parameter information............................................
                if len(param_prefix)-formatting_chars > prefix_width - 1:
                    output.append(param_prefix)
                    continue
                addon = param_prefix + ' #'
                first_addon = True
                addon_formatting = formatting_chars
            else:
                ## for subsequent lines space over prefix width........................................
                addon = (' ' * prefix_width) + '#'
                first_addon = False
                addon_formatting = 0
            ## if first word of description puts us over the screen width, bail........................
            if len(addon + description[0]) - addon_formatting + 1 > self.term_width:
                ## if we're doing the first line make sure it's output.................................
                if first_addon: output.append(addon)
                break
            while len(description) > 0:
                ## if the next description word puts us over break for the next line...................
                if len(addon + description[0]) - addon_formatting + 1 > self.term_width: break
                addon = addon + ' ' + description[0]
                description.pop(0)
            output.append(addon)
        out.write('\n'.join(output) + '\n')

    #--------- go functions -----------------------------------------------------------
    def __image_dflt( self, glb ):
        return [ ]

    def __image( self, glb ):
        if 'image' in glb: return glb['image']
        return [ ]

    def __image_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__image_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return [ ]
        description = 'Image(s) for which a mask cube will be created or updated'
        value = self.__image( self.__globals_( ) )
        (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'image': value},{'image': self.__schema['image']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('%-6.6s = %s%-23s%s' % ('image',pre,self.__to_string_(value),post),description,0+len(pre)+len(post))

    def __mask_dflt( self, glb ):
        return [  ]

    def __mask( self, glb ):
        if 'mask' in glb: return glb['mask']
        return [  ]

    def __mask_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__mask_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return [  ]
        description = 'Mask cubes(s) to be updated or created'
        value = self.__mask( self.__globals_( ) )
        (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'mask': value},{'mask': self.__schema['mask']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('%-6.6s = %s%-23s%s' % ('mask',pre,self.__to_string_(value),post),description,0+len(pre)+len(post))

    def __create_dflt( self, glb ):
        return True

    def __create( self, glb ):
        if 'create' in glb: return glb['create']
        return True

    def __create_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__create_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return True
        description = 'If equal to True, mask cubes are created if they do not exist'
        value = self.__create( self.__globals_( ) )
        (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'create': value},{'create': self.__schema['create']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('%-6.6s = %s%-23s%s' % ('create',pre,self.__to_string_(value),post),description,0+len(pre)+len(post))


    #--------- global default implementation-------------------------------------------
    @static_var('state', __sf__('casa_inp_go_state'))
    def set_global_defaults(self):
        self.set_global_defaults.state['last'] = self
        glb = self.__globals_( )
        if 'image' in glb: del glb['image']
        if 'mask' in glb: del glb['mask']
        if 'create' in glb: del glb['create']

    #--------- inp function -----------------------------------------------------------
    def inp(self):
        print("# createmask -- %s" % self._info_desc_)
        self.term_width, self.term_height = shutil.get_terminal_size(fallback=(80, 24))
        self.__image_inp( )
        self.__mask_inp( )
        self.__create_inp( )

    #--------- tget function ----------------------------------------------------------
    @static_var('state', __sf__('casa_inp_go_state'))
    def tget(self,savefile=None):
        from runpy import run_path
        filename = savefile
        if filename is None:
            filename = "createmask.last" if os.path.isfile("createmask.last") else "createmask.saved"
        if os.path.isfile(filename):
            glob = _find_frame( )
            newglob = run_path( filename, init_globals={ } )
            for i in newglob:
                glob[i] = newglob[i]
            self.tget.state['last'] = self
        else:
            print("could not find last file: %s\nsetting defaults instead..." % filename)
            self.set_global_defaults( )

    #--------- tput function ----------------------------------------------------------
    def tput(self,outfile=None):
        def noobj(s):
           if s.startswith('<') and s.endswith('>'):
               return "None"
           else:
               return s

        _postfile = outfile if outfile is not None else os.path.realpath('createmask.last')

        _invocation_parameters = OrderedDict( )
        _invocation_parameters['image'] = self.__image( self.__globals_( ) )
        _invocation_parameters['mask'] = self.__mask( self.__globals_( ) )
        _invocation_parameters['create'] = self.__create( self.__globals_( ) )

        try:
            with open(_postfile,'w') as _f:
                for _i in _invocation_parameters:
                    _f.write("%-20s = %s\n" % (_i,noobj(repr(_invocation_parameters[_i]))))
                _f.write("#createmask( ")
                count = 0
                for _i in _invocation_parameters:
                    _f.write("%s=%s" % (_i,noobj(repr(_invocation_parameters[_i]))))
                    count += 1
                    if count < len(_invocation_parameters): _f.write(",")
                _f.write(" )\n")
        except: return False
        return True

    def __call__( self, image=None, mask=None, create=None ):
        def noobj(s):
           if s.startswith('<') and s.endswith('>'):
               return "None"
           else:
               return s
        _prefile = os.path.realpath('createmask.pre')
        _postfile = os.path.realpath('createmask.last')
        task_result = None
        _arguments = [image,mask,create,]
        _invocation_parameters = OrderedDict( )
        if any(map(lambda x: x is not None,_arguments)):
            # invoke python style
            # set the non sub-parameters that are not None
            local_global = { }
            if image is not None: local_global['image'] = image
            if mask is not None: local_global['mask'] = mask
            if create is not None: local_global['create'] = create

            # the invocation parameters for the non-subparameters can now be set - this picks up those defaults
            _invocation_parameters['image'] = self.__image( local_global )
            _invocation_parameters['mask'] = self.__mask( local_global )
            _invocation_parameters['create'] = self.__create( local_global )

            # the sub-parameters can then be set. Use the supplied value if not None, else the function, which gets the appropriate default

        else:
            # invoke with inp/go semantics
            _invocation_parameters['image'] = self.__image( self.__globals_( ) )
            _invocation_parameters['mask'] = self.__mask( self.__globals_( ) )
            _invocation_parameters['create'] = self.__create( self.__globals_( ) )
        try:
            with open(_prefile,'w') as _f:
                for _i in _invocation_parameters:
                    _f.write("%-20s = %s\n" % (_i,noobj(repr(_invocation_parameters[_i]))))
                _f.write("#createmask( ")
                count = 0
                for _i in _invocation_parameters:
                    _f.write("%s=%s" % (_i,noobj(repr(_invocation_parameters[_i]))))
                    count += 1
                    if count < len(_invocation_parameters): _f.write(",")
                _f.write(" )\n")
        except: pass
        try:
            _logging_state_ = None
            assert _pc.validate(_invocation_parameters,self.__schema), create_error_string(_pc.errors)
            _logging_state_ = _start_log( 'createmask', [ 'image=' + repr(_pc.document['image']), 'mask=' + repr(_pc.document['mask']), 'create=' + repr(_pc.document['create']), ] )
            task_result = _createmask_t( image=_pc.document['image'],mask=_pc.document['mask'],create=_pc.document['create'], )
        except Exception as e:
            from traceback import format_exc
            from casatasks import casalog
            casalog.origin('createmask')
            casalog.post("Exception Reported: Error in createmask: %s" % str(e),'SEVERE')
            casalog.post(format_exc( ))
            raise #exception is now raised
            #task_result = False
        finally:
            try:
                os.rename(_prefile,_postfile)
            except: pass
            if _logging_state_:
                task_result = _end_log( _logging_state_, 'createmask', task_result )

        #Added if _createmask_t returns False and does not raise an exception.
        if task_result is False:
            raise

        return task_result #Still needed

createmask = _createmask( )
